/*
 * ImageCalibration
 *
 * Copyright (C) 2006 Guilhem Bonnefille <guilhem.bonnefille@gmail.com>
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <AppKit/AppKit.h>
#import "AppController.h"
#import "CalibratingImageModel.h"
#import "CalibratingPointModel.h"
#import "CalibratingImageView.h"
#import "GTGeoCoordinatesView.h"


@interface AppController (InternalMethods)
- (void) createPoint: (NSPoint) point;
- (void) selectAt: (NSPoint) point;
@end


@implementation AppController

- (id) init
{
  self = [super init];
  model = [[CalibratingImageModel alloc] init];

  creationEnabled = NO;

  return self;
}

- (void) dealloc
{
  [model release];
  [super dealloc];
}
- (void) enableCreation:(id)sender
{
  creationEnabled = YES;
}

- (void) disableCreation
{
  creationEnabled = NO;
}

- (void) loadImage: (id)sender
{
  NSOpenPanel *oPanel;
  int result;
  
  oPanel = [NSOpenPanel openPanel];
  [oPanel setAllowsMultipleSelection:NO];
  result = [oPanel runModalForDirectory: NSHomeDirectory()
		   file: nil
		   types: nil];
  
  if (result == NSOKButton)
    {
      NSEnumerator *filesToOpenEnumerator;
      NSString *theFilename;
      
      filesToOpenEnumerator = [[oPanel filenames] 
				objectEnumerator];
      theFilename = [filesToOpenEnumerator nextObject];
      
      if ( theFilename )
        {
          NSImage *image;
          
          image = [[NSImage alloc] 
                    initWithContentsOfFile: theFilename];
          
          [model setImage: image];
          [image release];
          
          [window setTitle: theFilename];
          [imageView setModel: model];  

          // Register as image position listener
          [imageView setImagePositionListener: self];
        }
    }
}

- (void) saveGeoData: (id)sender
{
  int ret;
  NSSavePanel* panel = [NSSavePanel savePanel];
  [panel setRequiredFileType: @"txt"];
  [panel setDirectory: [[NSFileManager defaultManager] currentDirectoryPath]];

  ret = [panel runModal];
  if (ret == NSFileHandlingPanelOKButton)
  {
    NSString *theFilename;
    
    theFilename = [panel filename];
      
    if ( theFilename ) {
      CalibratingPointModel *geoPoint;
      int i;
      FILE *file = NULL;
      const char *cFilename;

      cFilename = [theFilename cString];

      // Open the file
      file = fopen(cFilename, "w");

      if (file != NULL) {
        // Write a header/comment
        fprintf(file, "# x y long lat\n");
      
        for (i = 0 ; i < MAX_POINTS ; i++) {
          geoPoint = [model getPointAt: i];
          if (geoPoint != nil && [geoPoint isValid] == YES) {
            NSPoint imgCoord = [geoPoint point];
            t_latitude latitude = [geoPoint latitude];
            t_longitude longitude = [geoPoint longitude];
            fprintf(file, "%f %f %f %f\n",
                    imgCoord.x, imgCoord.y, longitude, latitude);
          }
        }

        fclose(file);
        file = NULL;
      }
    }
  }
}

// ImagePositionListener
- (void) mouseMovedOnImage: (NSPoint) point
{
  // Nothing for now
}

- (void) mouseDownOnImage: (NSPoint) point
{
  printf("mouseDownOnImage:%fx%f\n",
         point.x, point.y);

  if (creationEnabled == YES) {
    [self createPoint:point];
    [self disableCreation];
  } else {
    [self selectAt:point];
  }
}

- (void) createPoint: (NSPoint) point
{
  static futureIdx = 0;

  // Update model
  [[model getPointAt: futureIdx] setPoint: point];

  // Inform View
  [imageView setCoordinates: point forIndex:futureIdx];

  // Computing futureIdx
  futureIdx++;
  if (futureIdx >= MAX_POINTS)
    futureIdx = 0;
}

- (void) selectAt: (NSPoint) point
{
  BOOL selected = [imageView selectAt: point];
  if (selected) {
    CalibratingPointModel *point = [imageView selectedPoint];
    if (geoCoordView != nil) {
      [geoCoordView setLongitude: [point longitude]];
      [geoCoordView setLatitude: [point latitude]];
    }
  }
}

- (void) positionChanged:(id)view
{
  //NSLog(@"positionChanged");
  CalibratingPointModel *point = [imageView selectedPoint];
  if (point != nil && geoCoordView != nil) {
    [point setLongitude: [geoCoordView longitude]];
    [point setLatitude: [geoCoordView latitude]];
  }
  
}

@end
