/*
 * ImageCalibration
 *
 * Copyright (C) 2006 Guilhem Bonnefille <guilhem.bonnefille@gmail.com>
 * Inspired by Gautier Taravella on Sun Jun 20 2004.
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#import "CalibratingImageView.h"
#import "ImagePositionListener.h"
#import "CalibratingImageModel.h"
#import "CalibratingPointModel.h"
#import "CalibratingPointView.h"

#include <math.h>

@interface CalibratingImageView (InternalMethods)
- (NSPoint) imageCoordinates: (NSPoint) locationInWindow;
- (NSPoint) viewCoordinates: (NSPoint) locationInImage;

@end

@implementation CalibratingImageView

- (id) init
{
  int i;
  for (i = 0 ; i < MAX_POINTS ; ++i) {
    pointViews[i] = [CalibratingPointView alloc];
    [pointViews[i] initWithNumber: i + 1];
  }
  
  selectedPoint = -1;

  return self;
}

- (void) awakeFromNib {
  [self init];
  [self resetPoints];
}

- (void) resetPoints {
  int i;
  for (i = 0 ; i < MAX_POINTS ; i++) {
    [pointViews[i] resetPoint];
  }
}

- (void)drawRect:(NSRect)rect {
  NSLog(@"CalibratingPointView::drawRect begin");
    /*[[NSColor greenColor] set];
    NSRect frame = NSMakeRect(0,0,0,0);
    frame.size = [self frame].size;
    NSRectFill(rect);*/

    [super drawRect: rect]; 
    
    // Draw the points
    int i;
    for (i = 0 ; i < MAX_POINTS ; ++i) {
      NSPoint point = [pointViews[i] getPoint];
      NSPoint loc = [self viewCoordinates: point];
      [pointViews[i] draw: loc inRect: rect];
    }

  NSLog(@"CalibratingPointView::drawRect end");
}

- (void) setCoordinates: (NSPoint) coords forIndex: (int) idx {
  if (idx >= 0 && idx < MAX_POINTS) {
    BOOL needDisplay;
    // NSPoint point = [self viewCoordinates: coords];
    // needDisplay = [pointViews[idx] setPoint: point];
    needDisplay = [pointViews[idx] setPoint: coords];
    [self setNeedsDisplay: needDisplay];
  }
    
}

- (void)mouseDown:(NSEvent *)theEvent {
    NSPoint loc = [theEvent locationInWindow];
    NSPoint coords = [self imageCoordinates: loc];
    [[self window] makeFirstResponder: self];

    if (coords.x != -1 && listener != nil) {
        [listener mouseDownOnImage: coords];   
    }
}

- (void)mouseMoved:(NSEvent *)theEvent {
    NSPoint loc = [theEvent locationInWindow];
    NSPoint coords = [self imageCoordinates: loc];

    if (coords.x != -1 && listener != nil) {
        [listener mouseMovedOnImage: coords];   
    }
}

- (NSPoint) viewCoordinates: (NSPoint) loc {    
    NSSize viewSize = [self frame].size;
    
    float viewRatio = viewSize.width/viewSize.height;
    float imgRatio = (float)pixWide/(float)pixHigh;
    
    float disImgWidth, disImgHeight;
    float xOff, yOff;
    float scale;
    
    // Assume image is zoomed proportionally to fit
    if (imgRatio > viewRatio) {
        disImgWidth = viewSize.width;
        scale = disImgWidth / pixWide;
        disImgHeight = disImgWidth / imgRatio;
    } else {
        disImgHeight = viewSize.height;
        scale = disImgHeight / pixHigh;
        disImgWidth = imgRatio * disImgHeight;
    }
    
    // Assume image is centered in view
    xOff = (viewSize.width - disImgWidth) / 2;
    yOff = (viewSize.height - disImgHeight) / 2;
    
    float x = (scale * loc.x) + xOff;
    float y = viewSize.height - (scale * loc.y) - yOff;
    
    return NSMakePoint(x, y);
}

- (void) setFrame: (NSRect) frame {
    //NSLog(@"Setting frame to (%f,%f) %fx%f", frame.origin.x, frame.origin.y, frame.size.height, frame.size.width);
    [super setFrame: frame];
}

- (void) setFrameSize: (NSSize) size {
    //NSLog(@"Setting frame size to %fx%f", size.height, size.width);
    // Preserve aspect ratio
    float zoomX = size.width/(float)pixWide;
    float zoomY = size.height/(float)pixHigh;
    if (zoomX > zoomY)
        size.width = pixWide*zoomY;
    if (zoomY > zoomX)
        size.height = pixHigh*zoomX;
    
    // make sure size is an integer
    size.height = round(size.height);
    size.width = round(size.width);
    
    [super setFrameSize: size];
}

- (NSSize) imagePixelSize {
    return NSMakeSize(pixWide, pixHigh);   
}

- (void) setModel: (CalibratingImageModel*) newModel {
  [newModel retain];
  [model release];
  model = newModel;

  // Update the background image
  [self setImage: [model image]];

  // Update the point views
  int i;
  for (i = 0 ; i < MAX_POINTS ; ++i) {
    NSPoint coord = [[model getPointAt: i] point];
    // NSPoint point = [self viewCoordinates: coord];
    // [pointViews[i] setPoint: point];
    [pointViews[i] setPoint: coord];
  }
}

- (void) setImage: (NSImage *) image {
    if ([[image representations] count]) {
        // Get number of pixels in the image (the [image size] method
        // returns something that takes into account DPI)
        NSImageRep * rep = [[image representations] objectAtIndex:0];
        pixHigh = [rep pixelsHigh];
        pixWide = [rep pixelsWide];
        [super setImage: image];
    } else {
        // Error with the NSImage
        //NSException *exception = [NSException exceptionWithName:@"InvalidArgument"
        //                                                 reason:@"The passed image has no representation." userInfo:nil];
        //@throw exception;
    }

}

- (BOOL) acceptsFirstResponder {
    return YES;
}

- (NSPoint) imageCoordinates: (NSPoint) loc {
    loc = [self convertPoint:loc fromView:nil];
    
    NSSize viewSize = [self frame].size;
    
    float viewRatio = viewSize.width/viewSize.height;
    float imgRatio = (float)pixWide/(float)pixHigh;
    
    float disImgWidth, disImgHeight;
    float xOff, yOff;
    float scale;
    
    // Assume image is zoomed proportionally to fit
    if (imgRatio > viewRatio) {
        disImgWidth = viewSize.width;
        scale = disImgWidth / pixWide;
        disImgHeight = disImgWidth / imgRatio;
    } else {
        disImgHeight = viewSize.height;
        scale = disImgHeight / pixHigh;
        disImgWidth = imgRatio * disImgHeight;
    }
    
    // Assume image is centered in view
    xOff = (viewSize.width - disImgWidth) / 2;
    yOff = (viewSize.height - disImgHeight) / 2;
    
    float x = (loc.x - xOff)/scale;
    float y = ((viewSize.height - loc.y) - yOff)/scale;    // Cocoa views are bottom up
    
    //NSLog(@"within image (%fx%f) viewed: %f  , %f",imgSize.width, imgSize.height, loc.x - xOff, (viewSize.height - loc.y) - yOff);
    
    if (x < 0 || x > pixWide || y < 0 || y > pixHigh) {
        x = y = -1;
    }
    
    //NSLog(@"scale: %f imgRatio:%f viewRatio:%f xOff:%f yOff:%f", scale, imgRatio, viewRatio, xOff, yOff);
    return NSMakePoint(x, y);
}


// Selection

- (void) setImagePositionListener: (id) newListener {
    listener = newListener;   
}

- (void) clearSelection
{
  int i;
  for (i = 0 ; i < MAX_POINTS ; i++) {
    [pointViews[i] unselect];
  }

  if (selectedPoint != -1)
    // There was a selection
    // We will need to redraw
    [self setNeedsDisplay: YES];
  selectedPoint = -1;
}

- (BOOL) selectAt: (NSPoint) loc
{
  int i;
  [self clearSelection];
  for (i = 0 ; i < MAX_POINTS ; i++) {
    NSPoint viewLoc = [self viewCoordinates: loc];
    BOOL selected = [pointViews[i] selectAt: viewLoc];
    if (selected) {
      selectedPoint = i;
      [pointViews[i] select];
      [self setNeedsDisplay: YES];
     return YES;
    }
  }
  return NO;
}

- (CalibratingPointModel*) selectedPoint
{
  if (selectedPoint >= 0)
    return [model getPointAt: selectedPoint];
  else
    return nil;
}

@end
