//
//  ImageCalibrationController.m
//  FlightTrack
//
//  Created by Gautier Taravella on Sun Jun 20 2004.
//  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
//

#import "ImageCalibrationController.h"
#import "CalibratingImageView.h"
#import "CenteringClipView.h"
#import "GTGeoCoordinatesFormatter.h"
#include "gdal.h"
#include <gdal_priv.h>
#include "ogr_srs_api.h"
#include "cpl_string.h"
#include <ogr_spatialref.h>
#include "GDALUtils.h"

#define ISNAN(a) (!((a) == (a)))

#define NONE_STRING @"None"
#define UTM_STRING @"Universal transverse mercator"

#define TAB_NONE 0
#define TAB_OTHERS 1
#define TAB_UTM 2

#define GEO_REF_2_OR_3_POINTS 0
#define GEO_REF_MANUAL 1

static const char * GEOREF_METHODS[] = {"GCPS", "MANUAL"}; 

#define TRANSFORM_A 1 
#define TRANSFORM_B 2 
#define TRANSFORM_C 0 
#define TRANSFORM_D 4 
#define TRANSFORM_E 5 
#define TRANSFORM_F 3


#include <algorithm>

using namespace std;

static const int NUM_DATUMS = 4;
static const char * DATUMS[] = {"WGS84", "WGS72", "NAD83", "NAD27", NULL };

static bool isGeoTransformValid (double * transform) {
    bool valid = true;
    for (int i = 0 ; i < 6 ; ++i) {
        if (ISNAN(transform[i])) {
            valid = false;
            break;
        }
    }
    
    return valid;
}

static void updateGCPLatLon (GDAL_GCP * pGCP, int count, double * transform, OGRCoordinateTransformation * proj) {
    for (int i = 0; i < count; ++i) {
        
        if (pGCP->dfGCPPixel != -1 && pGCP->dfGCPLine != -1) {
            GDALApplyGeoTransform(transform, pGCP->dfGCPPixel, pGCP->dfGCPLine, &(pGCP->dfGCPX), &(pGCP->dfGCPY));
            if (proj != NULL) {
                if (proj->Transform(1, &(pGCP->dfGCPX), &(pGCP->dfGCPY)) != 0) {
                    NSLog(@"%s:%d Transform error on GCP", __FILE__, __LINE__);
                }
            }
        }
        pGCP++;
    }
}

@interface ImageCalibrationController (InternalMethods)
- (void) updateGeoRefState;
- (void) updateGeoRefView;
- (void) resetValues;
- (void) writeCalibration;
- (void) refreshProjection;

- (void) updateDoubleArray: (double *) values fromNSForm: (NSForm *) form;
- (void) updateNSForm: (NSForm *) form fromDoubleArray: (double *) values;

@end

@implementation ImageCalibrationController

- (id)init {
    reverseTrans = NULL;
    forwardTrans = NULL;
    srs = NULL;
    
    // Create array of geographic coordinate systems
    paGeoCS = new (OGRSpatialReference *) [NUM_DATUMS+1];
    paGeoCS[NUM_DATUMS] = NULL;
    for (int i = 0; i < NUM_DATUMS; ++i) {
        paGeoCS[i] = new OGRSpatialReference();
        paGeoCS[i]->SetWellKnownGeogCS(DATUMS[i]);
    }
    
    if (window == nil) {
        [NSBundle loadNibNamed: @"ImageGeoreference" owner: self];
        id docView = [[scrollView documentView] retain];
        id newClipView = [[CenteringClipView alloc] initWithFrame:[[scrollView contentView] frame]];
        [newClipView setBackgroundColor:[NSColor windowBackgroundColor]];
        if ([NSCursor respondsToSelector:@selector(crosshairCursor)])
            [newClipView setDocumentCursor:[NSCursor crosshairCursor]];
        [scrollView setContentView:(NSClipView *)newClipView];
        [newClipView release];
        [scrollView setDocumentView:docView];
        [scrollView setPostsFrameChangedNotifications:YES];
        [[NSNotificationCenter defaultCenter]  addObserver:self
                                                  selector:@selector(scrollViewFrameChanged:)
                                                      name:@"NSViewFrameDidChangeNotification"
                                                    object:scrollView ];
        [docView release];
    }
    [view setEditable:NO];  // The user should not drag and drop on the view
    
    return self;
}

- (void) dealloc {
}

- (void) awakeFromNib {
    [cursorGeoTF setEditable:NO];
    [cursorGeoTF setBezeled:NO];

    [self scrollViewFrameChanged: nil];
    
    geoPts[0] = latLon1V;
    geoPts[1] = latLon2V;
    geoPts[2] = latLon3V;
    
    [self action: self];
    
    // Fill popup menu with datums
    [datumPopup removeAllItems];
    for (int i = 0; i < NUM_DATUMS; ++i) {
        [datumPopup addItemWithTitle: [NSString stringWithCString:DATUMS[i]]];   
    }
    
    // Create array containing projection methods
    NSMutableArray * popupList = [NSMutableArray arrayWithCapacity:100];
    
    // Add 2 special items
    [popupList addObject:NONE_STRING];
    [popupList addObject:UTM_STRING];
    
    // Add other projection names
    char ** projList = OPTGetProjectionMethods();
    for (int i = 0; projList[i] != NULL; ++i) {
        // Convert '_' to spaces
        char * s = projList[i];
        while ((s = strchr(s, '_')) != NULL) s[0] = ' '; 
        [popupList addObject: [NSString stringWithCString:projList[i]]];
    }
    CSLDestroy(projList);
    
    // Sort array
    [popupList sortUsingSelector:@selector(compare:)];
    
    [projPopup addItemsWithTitles: popupList];
}

- (void)scrollViewFrameChanged: (NSNotification *) notification {
    [scrollView removeTrackingRect: trackingRect];
    trackingRect = [scrollView addTrackingRect:[scrollView convertRect:[scrollView frame] toView:[window contentView]]
                                         owner:self userData:nil assumeInside:false];
}

- (void)mouseEntered:(NSEvent *)theEvent {
    [window makeFirstResponder:view];
    [window setAcceptsMouseMovedEvents: YES];
}


- (void)mouseExited:(NSEvent *)theEvent {
    [cursorXTF setStringValue: @"-"];
    [cursorYTF setStringValue: @"-"];
    [cursorGeoTF setLatitude:NAN];
    [cursorGeoTF setLongitude:NAN];
    [window setAcceptsMouseMovedEvents: NO];
}

- (NSString *) fileToCalibrate {
    return fileName;
}

- (void) setFileToCalibrate: (NSString *) fileToReference {
    // TODO: Check that no file is being referenced and 
    // display a confirmation if so.
    
    [fileToReference retain];
    [fileName release];
    fileName = fileToReference;
    
    [self resetValues];
    [view setImage:nil];
    [window setAcceptsMouseMovedEvents:YES];
    
    NSImage * image = nil;
    
    
    GDALDataset *ds = (GDALDataset *) GDALOpen([fileName UTF8String], GA_ReadOnly);
    if (ds == NULL) {
        NSLog(@"Could not open file to calibrate with GDAL: %@", fileName);
        return;
    }
    
    // Try to open the original file with NSImage methods (if we can't we will use GDAL to create a BitmapRep)
    if ([[fileName pathExtension] caseInsensitiveCompare:@"vrt"] == NSOrderedSame) {
        image = [[NSImage alloc] initWithContentsOfFile:[fileName stringByDeletingPathExtension]];
        if (image == nil) NSLog(@"Can't create image with original (not VRT) file");
    }
    
    int bandsMatrix[5]; // Maps the index of the bands in the apple color space to the number of the band in the GDALDataset
    
    if (image == nil) {
        if (ds->GetRasterCount() < 1 || ds->GetRasterCount() > 4) {
            NSLog(@"Can't interpret number of bands");
        } else {
            bool hasColorPalette = ds->GetRasterBand(1)->GetColorInterpretation() == GCI_PaletteIndex;
            
            // If the image has a color palette, check that it is properly filled.
            // Some maps downloaded from NIMA raster roam don't have a valid color palette
            if (hasColorPalette) {
                GDALColorTable * palette = ds->GetRasterBand(1)->GetColorTable();
                if (palette->GetColorEntryCount() > 256 || palette->GetColorEntryCount() < 1) {
                    // TODO: abort overview process
                }
                
                // Test if all color entries are equal
                const GDALColorEntry * firstEntry = palette->GetColorEntry(0);
                const GDALColorEntry * entry;
                bool allTheSameColor = true;
                for (int i = 1; i < palette->GetColorEntryCount(); ++i) {
                    entry = palette->GetColorEntry(i);
                    if (entry != firstEntry) allTheSameColor = false;
                    if (!allTheSameColor) break;
                }
                
                if (allTheSameColor) {
                    NSLog(@"Invalid color palette, assuming grayscale");
                    hasColorPalette = false;
                }
                
                // Test if the image is grayscale (all color components (except alpha) are equal)
                bool grayColor = true;
                for (int i = 0; i < palette->GetColorEntryCount(); ++i) {
                    entry = palette->GetColorEntry(i);
                    if (entry->c1 != entry->c2 || entry->c2 != entry->c3) grayColor = false;
                    if (!grayColor) break;
                }
                
                if (grayColor) {
                    NSLog(@"Gray color palette, will convert to 1 band grayscale");
                    hasColorPalette = false;
                }
                
            }
            
            // Create Image and ImageRep
            int numBands = ds->GetRasterCount();
            int tgtNumBands = numBands;
            if (hasColorPalette) tgtNumBands = 3;
            bool hasAlpha = false;
            
            NSString * cs = (tgtNumBands < 3) ? NSCalibratedWhiteColorSpace : NSCalibratedRGBColorSpace;
            
            //NSLog(@"Init Bitmap rep");
            
            
            NSBitmapImageRep * imageRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL 
                                                                                  pixelsWide:ds->GetRasterXSize()
                                                                                  pixelsHigh:ds->GetRasterYSize()
                                                                               bitsPerSample:8
                                                                             samplesPerPixel:ds->GetRasterCount()
                                                                                    hasAlpha:hasAlpha
                                                                                    isPlanar: YES
                                                                              colorSpaceName:cs
                                                                                 bytesPerRow:0
                                                                                bitsPerPixel:0];
            
            //NSLog(@"Alloc image");
            
            image = [[NSImage alloc] initWithSize:NSMakeSize(ds->GetRasterXSize(), ds->GetRasterYSize())];
            
            //NSLog(@"Add Bitmap rep");
            
            [image addRepresentation: imageRep];
            
            
            //NSLog(@"Read bands");
            
            // Copy GDAL overview data in the data buffers provided by the ImageRep
            unsigned char * planes[5];
            [imageRep getBitmapDataPlanes: (unsigned char **) &planes];
            for (int band = 1 ; band <= numBands ; ++band) {
                ds->GetRasterBand(band)->RasterIO(GF_Read, 0, 0, ds->GetRasterXSize(), ds->GetRasterYSize(),
                                                  planes[band-1], ds->GetRasterXSize(), ds->GetRasterYSize(), GDT_Byte, 0, 0);
            }
            
            // If color indexed, convert to RGB.
            if (hasColorPalette) {
                //NSLog(@"do color palette");
                
                const GDALColorTable * palette = ds->GetRasterBand(1)->GetColorTable();
                const GDALColorEntry * entry;
                unsigned char * colorIndexPlane = planes[0];
                for (int i = 0 ; i < ds->GetRasterXSize()*ds->GetRasterYSize() ; ++i) {
                    entry = palette->GetColorEntry(colorIndexPlane[i]);
                    planes[0][i] = entry->c1;
                    planes[1][i] = entry->c2;
                    planes[2][i] = entry->c3;
                }
            } 
            
            //NSLog(@"Done");
            
        }
    }
    
    
    if (image == nil) {
        NSRunAlertPanel(@"The map can't be opened. The image format might be unsupported or an error occured.",
                        @"", @"OK", nil, nil);
        [fileName release];
        fileName = nil;
        return;
    }
    
    
    // Try to open previous image projection
    OGRSpatialReference * readSrs = new OGRSpatialReference(ds->GetProjectionRef());
    if (readSrs) {
        // -------- Update HMI with SRS ---------
        
        // What is the GeoCS ?
        int idx;
        for (idx = 0; idx < NUM_DATUMS; ++idx) {
            if (readSrs->IsSameGeogCS(paGeoCS[idx]))
                break;
        }
        
        if (idx < NUM_DATUMS) {
            //NSLog(@"Using DATUM %s", DATUMS[idx]);
            [datumPopup selectItemWithTitle:[NSString stringWithCString: DATUMS[idx]]];
        } else NSLog(@"Could not find datum");
        
        if (readSrs->IsProjected()) {
            // Convert '_' to spaces and select projection in menu
            const char * s = readSrs->GetAttrValue("PROJECTION");
            char * proj = new char[strlen(s)+1];
            strcpy(proj, s);
            s = proj;
            while ((proj = strchr(proj, '_')) != NULL) proj[0] = ' '; 
            [projPopup selectItemWithTitle: [NSString stringWithCString:s]];
            delete [] proj;
            [self projAction: projPopup];
            
            // Populate projection parameters
            NSFormCell * cell;
            double d;
            OGRErr err;
            for (int i = 0 ; (cell = [projParamMx cellAtIndex: i]) != nil ; ++i) {
                d = readSrs->GetProjParm([[cell title] UTF8String], NAN, &err);
                //NSLog(@"Input parm %@, %f, %d", [cell title], d, err);
                if (err == OGRERR_NONE) 
                    if (!ISNAN(d)) {
                        //NSLog(@"setting value");
                        [cell setDoubleValue: d];
                        //NSLog(@"set value %f from cell %f", d, [cell doubleValue]);
                        
                    }
                        else NSLog(@"Could not fetch projection parameter for %@", [cell title]);
            }
        } // END if(readSrs->IsProjected())
    } // END if (readSrs)
    
    [self refreshProjection];
        
    // Copy MAX_CALIBRATION_POINTS GCPs to internal state
    // FIXME: GCP Spatial reference is not considered, maybe it should, if the DS was not created with FlightTrack
    const GDAL_GCP * gcpds = ds->GetGCPs();
    for (int i = 0; i < min(ds->GetGCPCount(), MAX_CALIBRATION_POINTS) ; ++i) {
        gcps[i] = gcpds[i];
        [view setCoordinates: NSMakePoint(gcps[i].dfGCPPixel, gcps[i].dfGCPLine) forIndex:i];
    }
    
    // Default geo-reference method for an image with no information
    geoRefMethod = GEO_REF_2_OR_3_POINTS; 
    
    // Try to read previously used geo-reference method
    const char * method = ds->GetMetadataItem("FlightTrackGeoRefMethod");
    for (int i = 0; method != NULL && i <= GEO_REF_MANUAL ; ++i) {
        if (strcmp(method, GEOREF_METHODS[i]) == 0) {
            geoRefMethod = i;
            break;
        }
    }
        
    // Set geo transform only if there was no failure (i.e. DS has transform)
    double trans[6];
    if (ds->GetGeoTransform(trans) == CE_None) {
        memcpy (geoTransform, trans, 6*sizeof(double));
        isGeoreferenced = isGeoTransformValid(geoTransform);
    }
    
    [self updateGeoRefView];
    
    [view setImage:image];
    [image release];
    [view setFrameOrigin:NSMakePoint(0,0)];
    [view setFrameSize:[[scrollView contentView] frame].size];
    [window makeKeyAndOrderFront: self];
    [window makeFirstResponder:view];
    [window setAcceptsMouseMovedEvents: YES];
}

- (void) resetValues {
    // HMI
    [view resetPoints];
    [cursorXTF setStringValue: @"-"];
    [cursorYTF setStringValue: @"-"];
    [cursorGeoTF setLatitude:NAN];
    [cursorGeoTF setLongitude:NAN];
    
    // Reset state
    for (int i = 0; i < MAX_CALIBRATION_POINTS; i++) {
        gcps[i].dfGCPX = NAN;
        gcps[i].dfGCPY = NAN;
        gcps[i].dfGCPPixel = -1;
        gcps[i].dfGCPLine = -1;
    }
    geoRefMethod = GEO_REF_2_OR_3_POINTS;
    for (int i = 0 ; i < 6 ; i++) {
        geoTransform[i] = NAN;
    }
    isGeoreferenced = NO;
    [self updateGeoRefView]; 
    
    
    // Reset projection & datum
    [ogrProjName release];
    ogrProjName = nil;
    [datumPopup selectItemWithTitle:@"WGS84"];
    [projPopup selectItemWithTitle:NONE_STRING];
    [projTabView selectTabViewItemAtIndex:TAB_NONE];
    [self refreshProjection];
}


- (IBAction) action: (id) sender {
    NSSize maxSize = [view imagePixelSize];
    NSSize newSize = [view frame].size;
    
    if (sender == zoomInB) {
        newSize.width *= 2;
        newSize.height *= 2;
        if (newSize.width >= maxSize.width && newSize.height >= maxSize.height) {
            newSize = maxSize;
        }
        [view setFrameSize: newSize];
        [view setNeedsDisplay:YES];
    } else if (sender == zoomOutB) {
        newSize.width /= 2;
        newSize.height /= 2;
        [view setFrameSize: newSize];
        [view setNeedsDisplay:YES];
    } else if (sender == zoomToFitB) {
        [view setFrameSize:[[scrollView contentView] frame].size];
    } else if (sender == actualSizeB) {
        [view setFrameSize: maxSize];
        [view setNeedsDisplay:YES];
    } else if (sender == addMapB) {
        // Look for an existing calibration file (.wld)
        NSString * wldFile = [fileName stringByDeletingPathExtension];
        wldFile = [wldFile stringByAppendingPathExtension:@"wld"];
        
        if ([[NSFileManager defaultManager] fileExistsAtPath:wldFile]) {
            NSBeginAlertSheet(nil, @"Yes", @"No", nil, window, self, @selector(sheetDidEnd:returnCode:contextInfo:), nil, nil, @"The map is already calibrated. Are you sure you want to overwrite the previous calibration?");
        } else {
            [self writeCalibration];
        }
    }
    
    
}

- (IBAction) geoRefAction: (id) sender {
    if (sender == selMethodMx) {
        geoRefMethod = [[selMethodMx selectedCell] tag];
        //NSLog(@"New geo ref method: %d", geoRefMethod);        
    }
    
    [self updateGeoRefState];
}

- (IBAction) projAction: (id) sender {
    if (sender == projPopup) {
        //NSLog(@"Projection selected: %@", [sender titleOfSelectedItem]);
        NSString * proj = [sender titleOfSelectedItem];
        
        if ([proj isEqualToString:NONE_STRING]) {
            [projTabView selectTabViewItemAtIndex: TAB_NONE];
        } else if ([proj isEqualToString:UTM_STRING]) {
            [projTabView selectTabViewItemAtIndex: TAB_UTM];
            [projUTMZoneTF setStringValue:@""];
            [projUTMHemiMx selectCellAtRow:0 column:0];
        } else {
            [projTabView selectTabViewItemAtIndex: TAB_OTHERS];
            
            // Empty parameters matrix (could'nt find a simpler way to do that...)
            while ([projParamMx cellAtRow:0 column:0] != nil) {
                [projParamMx removeEntryAtIndex: 0];
            }
            
            // Convert ' ' to '_' and get list of parameters 
            const char * projs = [proj UTF8String];
            char * ogrStr = new char [strlen(projs) + 1];
            strcpy(ogrStr, projs);
            char * s = ogrStr;
            while ((s = strchr(s, ' ')) != NULL) s[0] = '_';
            char ** params = OPTGetParameterList(ogrStr, NULL);
            
            // Keep OGR projection name for future use
            [ogrProjName release];
            ogrProjName = [[NSString stringWithCString: ogrStr] retain];
            delete ogrStr;
            
            // Add entry for each parameter of the projection
            NSFormCell * cell;
            for (int i = 0; params[i] != NULL; ++i) {
                
                // Don't add false easting and false northings parameters in HMI
                if (1 || strcmp(params[i], SRS_PP_FALSE_EASTING) && strcmp(params[i], SRS_PP_FALSE_NORTHING)) {
                    // Add entry
                    cell = [projParamMx addEntry:[NSString stringWithCString:params[i]]];
                    
                    // Set appropriate formatters for latitude and longitude
                    if (!strcmp(params[i], SRS_PP_CENTRAL_MERIDIAN) || !strcmp(params[i], SRS_PP_LONGITUDE_OF_CENTER) 
                        || !strcmp(params[i], SRS_PP_LONGITUDE_OF_ORIGIN) || !strcmp(params[i], SRS_PP_LONGITUDE_OF_POINT_1) 
                        || !strcmp(params[i], SRS_PP_LONGITUDE_OF_POINT_2) || !strcmp(params[i], SRS_PP_LONGITUDE_OF_POINT_3)) {
                        [cell setFormatter: [GTGeoCoordinatesFormatter applicationLongitudeFormatter]];
                        [cell setDoubleValue:INFINITY];
                    } else if (!strcmp(params[i], SRS_PP_STANDARD_PARALLEL_1) || !strcmp(params[i], SRS_PP_STANDARD_PARALLEL_2) 
                               || !strcmp(params[i], SRS_PP_PSEUDO_STD_PARALLEL_1) || !strcmp(params[i], SRS_PP_LATITUDE_OF_CENTER) 
                               || !strcmp(params[i], SRS_PP_LATITUDE_OF_ORIGIN) || !strcmp(params[i], SRS_PP_LATITUDE_OF_POINT_1)
                               || !strcmp(params[i], SRS_PP_LATITUDE_OF_POINT_2) || !strcmp(params[i], SRS_PP_LATITUDE_OF_POINT_3)) {
                        [cell setFormatter: [GTGeoCoordinatesFormatter applicationLatitudeFormatter]];
                        [cell setDoubleValue:INFINITY];
                    }
                }
                
            }
            
            // Matrix has been modified
            [projParamMx sizeToCells];
            CSLDestroy(params);
        }
    }
    
    [self refreshProjection];
}


- (void) refreshProjection {
    int ret;
    
    // Delete SRS and recreate new one
    if (srs) delete srs;
    srs = new OGRSpatialReference();
    
    // Set datum
    srs->SetWellKnownGeogCS([[datumPopup title] UTF8String]);
    
    if ([[projPopup title] isEqualToString: NONE_STRING]) {
        // No projection
    } else if ([[projPopup title] isEqualToString: UTM_STRING]) {
        // Set UTM zone
        if ([projUTMZoneTF intValue] != 0) {
            srs->SetUTM([projUTMZoneTF intValue], [[projUTMHemiMx selectedCell] tag] == 0);
        }
    } else {
        // Set projection method
        srs->SetProjection([ogrProjName UTF8String]);
        
        // Set projection parameters
        NSFormCell * cell;
        for (int i = 0 ; (cell = [projParamMx cellAtIndex: i]) != nil ; ++i) {
            //NSLog(@"cell string:%@ doubme:%f", [cell stringValue], [cell doubleValue]);
            if ([[cell stringValue] length] != 0 && !ISNAN([cell doubleValue]) && [cell doubleValue] != INFINITY) {
                //NSLog(@"setting proj parm %@: %f", [cell title], [cell doubleValue]);
                ret = srs->SetProjParm([[cell title] UTF8String], [cell doubleValue]);
                if (ret != OGRERR_NONE) NSLog(@"Error (%d) while setting projection (%@) attribute %@ to %f", ret, ogrProjName,
                                              [cell title], [cell doubleValue]);
            }
        }
    }
    
    // Try to validate SRS
    ret = srs->Validate();
    if (ret != OGRERR_NONE) {
        switch (ret) {
            case OGRERR_CORRUPT_DATA:
                NSLog(@"Error while validating SRS: OGRERR_CORRUPT_DATA");
                break;
            case OGRERR_UNSUPPORTED_SRS:
                NSLog(@"Error while validating SRS: OGRERR_UNSUPPORTED_SRS");
                break;
            default:
                NSLog(@"Unknown error while validating SRS: %d", ret);
                break;
        }
    }
    
    // Create transformation object
    if (reverseTrans) delete reverseTrans;
    if (forwardTrans) delete forwardTrans;
    OGRSpatialReference * latLonSrs = srs->CloneGeogCS(); // Convert to selected datum so that the user can check that projection is OK on the map.
    reverseTrans = OGRCreateCoordinateTransformation(srs, latLonSrs);
    forwardTrans = OGRCreateCoordinateTransformation(latLonSrs, srs);
    if (forwardTrans == NULL) {
        NSLog(@"Could not create forward coordinate transformation object");
    }
    if (reverseTrans == NULL) {
        NSLog(@"Could not create reverse coordinate transformation object");
    }
    
    char * wkt;
    srs->exportToPrettyWkt(&wkt);
    //NSLog(@"SRS:\n%s", wkt);
    OGRFree(wkt);
    
    [self updateGeoRefState];
}

- (void) writeCalibration {
    // Everything should work since it's a VRT dataset
    GDALDataset *poVRTDS = (GDALDataset *) GDALOpen([fileName UTF8String], GA_Update);
        
    // Save Geo Transform
    poVRTDS->SetGeoTransform(geoTransform);
    
    // Save GCPs and projection
    char * wktProj = NULL;
    char * wktDatum = NULL;
    
    if (srs != NULL) {
        srs->exportToWkt(&wktProj);
        OGRSpatialReference * geoCS = srs->CloneGeogCS();
        geoCS->exportToWkt(&wktDatum);
        delete geoCS;
    }
    poVRTDS->SetProjection(wktProj);
    poVRTDS->SetGCPs(MAX_CALIBRATION_POINTS, gcps, wktDatum ? wktDatum : "");
    
    if (wktProj != NULL) OGRFree(wktProj);
    if (wktDatum != NULL) OGRFree(wktDatum);
    
    // Save geoRef method
    poVRTDS->SetMetadataItem("FlightTrackGeoRefMethod", GEOREF_METHODS[geoRefMethod]);
    
    delete poVRTDS;
    
    // Notify
    [[NSNotificationCenter defaultCenter] postNotificationName:NOTIFICATION_IMAGE_CALIBRATED object:self];
}

/** updates an array of double with the data in a form.
* @param form the tag f each cell shall match an index of the array.
* @param values an array of double into which the parsed values will be put.
* @todo be more robust to non empty string that is not a double
*/
- (void) updateDoubleArray: (double *) values fromNSForm: (NSForm *) form {
    NSFormCell * cell;
    int tag;
    // Iterate on all cells of the form
    for (int i = 0 ; (cell = [form cellAtIndex: i]) != nil ; ++i) {
        tag = [cell tag];
        if ([[cell stringValue] length] != 0 && !ISNAN([cell doubleValue]) && [cell doubleValue] != INFINITY) {
            values[tag] = [cell doubleValue];
        } else {
            values[tag] = NAN;
        }
    }   
}

/** Updates an NSForm from an array of double.
* @param form the tag f each cell shall match an index of the array.
* @param values an array of double into which the values to put in the form will be read.
* @todo be more robust to non empty string that is not a double
*/
- (void) updateNSForm: (NSForm *) form fromDoubleArray: (double *) values{
    NSFormCell * cell;
    int tag;
    // Iterate on all cells of the form
    for (int i = 0 ; (cell = [form cellAtIndex: i]) != nil ; ++i) {
        tag = [cell tag];
        if (ISNAN(values[tag])) {
            [cell setStringValue:@""];
        } else {
            [cell setFloatValue:values[tag]];
            
        }
    }   
}


- (void) updateGeoRefState {
    
    //NSLog (@"Update of geo ref state with method %d ...", geoRefMethod);
    
    if (geoRefMethod == GEO_REF_2_OR_3_POINTS) {
        // Update state with the entered 3 points lat/lon
        for (int i = 0; i < MAX_CALIBRATION_POINTS; i++) {
            gcps[i].dfGCPX = [geoPts[i] longitude];
            gcps[i].dfGCPY = [geoPts[i] latitude];
        }
        
        // Select valid GCPs and convert them to projected points if applicable
        int validGCPCount = 0;        
        for (int i = 0; i < MAX_CALIBRATION_POINTS; i++) {
            if (!ISNAN(gcps[i].dfGCPX) && !ISNAN(gcps[i].dfGCPY) && gcps[i].dfGCPPixel != -1 && gcps[i].dfGCPLine != -1) {
                gcpsForComputation[validGCPCount] = gcps[i];
                
                // Project entered coordinates using projection system
                if (forwardTrans != NULL) {
                    if (!forwardTrans->Transform(1, &(gcpsForComputation[validGCPCount].dfGCPX), 
                                                &(gcpsForComputation[validGCPCount].dfGCPY))) {
                        NSLog(@"Transform error on GCP");
                    }
                }
                ++validGCPCount;
            }
        }
        
        if (validGCPCount > 1) {
            // Perform GCP fit
            if (GDALGCPsToGeoTransform(validGCPCount, gcpsForComputation, geoTransform, true)) {
                isGeoreferenced = YES;
            }  else {
                isGeoreferenced = NO;
            }
        } else if (validGCPCount == 1) {
            // Check for zero scales (not allowed)
            if (geoTransform[TRANSFORM_A] == 0)
                geoTransform[TRANSFORM_A] = NAN;
            if (geoTransform[TRANSFORM_E] == 0)
                geoTransform[TRANSFORM_E] = NAN;
        
            // Compute C and F from entered point by considering 
            // they are the worldX and worldY and converting the
            // opposite of the corresponding pixel offset
            geoTransform[TRANSFORM_C] = gcpsForComputation[0].dfGCPX;
            geoTransform[TRANSFORM_F] = gcpsForComputation[0].dfGCPY;
        
            if (isGeoTransformValid(geoTransform)) {
                // Apply geotransform
                double worldX, worldY;
                GDALApplyGeoTransform(geoTransform, -gcpsForComputation[0].dfGCPPixel, -gcpsForComputation[0].dfGCPLine, &worldX, &worldY);
                geoTransform[TRANSFORM_C] = worldX;
                geoTransform[TRANSFORM_F] = worldY;
            } else {
                geoTransform[TRANSFORM_C] = NAN;
                geoTransform[TRANSFORM_F] = NAN;
            }
        
            // Check that all coefficients are valid
            isGeoreferenced = isGeoTransformValid(geoTransform);
        }
        
    } else if (geoRefMethod == GEO_REF_MANUAL) {
        // Update all geo transform
        [self updateDoubleArray: geoTransform fromNSForm: affineTransformMx];
            
        // Check for zero scales (not allowed)
        if (geoTransform[TRANSFORM_A] == 0)
            geoTransform[TRANSFORM_A] = NAN;
        if (geoTransform[TRANSFORM_E] == 0)
            geoTransform[TRANSFORM_E] = NAN;
        
        // Check that all coefficients are valid
        isGeoreferenced = isGeoTransformValid(geoTransform);
    }
    
    //NSLog (@"Update of geo ref state successfull: %d", isGeoreferenced);
    
    [self updateGeoRefView];
}

- (void) updateGeoRefView {
    NSString * unitScale =  (srs != NULL && srs->IsProjected()) ? @"m/pix" : @"/pix";
    NSString * refScale =  (srs != NULL && srs->IsProjected()) ? @"m" : @"";
    
    // Widgets state
    NSFormCell * cell;
    // Iterate on all cells of the form
    for (int i = 0 ; (cell = [affineTransformMx cellAtIndex: i]) != nil ; ++i) {
        [cell setEditable:geoRefMethod == GEO_REF_MANUAL];
    }
    [affineTransformMx setNeedsDisplay:YES];
    
    for (int i = 0 ; i < MAX_CALIBRATION_POINTS; ++i) {
        [geoPts[i] setEditable: geoRefMethod != GEO_REF_MANUAL];
    }
    
    [geoRefComputeImageV setImage:[NSImage imageNamed: (geoRefMethod == GEO_REF_MANUAL ? @"computation-no" : @"computation-yes")]];

    // Popup menu and tab panel
    [selMethodMx selectCellWithTag:geoRefMethod];   
    
    // Method 2 or 3 points widgets
    for (int i = 0; i < MAX_CALIBRATION_POINTS; i++) {
        [geoPts[i] setLatitude: gcps[i].dfGCPY];
        [geoPts[i] setLongitude: gcps[i].dfGCPX];
    }
    
    // Method manual
    [self updateNSForm: affineTransformMx fromDoubleArray: geoTransform];
    [[affineTransformUnitsMx cellAtRow:0 column:0] setTitle:unitScale];
    [[affineTransformUnitsMx cellAtRow:1 column:0] setTitle:unitScale];
    [[affineTransformUnitsMx cellAtRow:2 column:0] setTitle:refScale];
    [[affineTransformUnitsMx cellAtRow:3 column:0] setTitle:unitScale];
    [[affineTransformUnitsMx cellAtRow:4 column:0] setTitle:unitScale];
    [[affineTransformUnitsMx cellAtRow:5 column:0] setTitle:refScale];
    
    // General for Geo Ref
    [geoRefStatusImageV setImage:[NSImage imageNamed: (isGeoreferenced ? @"OK" : @"NOK")]];
    [addMapB setEnabled: isGeoreferenced];
    
}


- (void) positionChanged: (GTGeoCoordinatesView *) view {
    [self updateGeoRefState];
}

- (void) mouseMovedOnImage: (NSPoint) point {
    [cursorXTF setIntValue: (int) point.x];
    [cursorYTF setIntValue: (int) point.y];
    
    // Display geographic coordinates
    if (isGeoreferenced) {
        double x = point.x, y = point.y;
        image2LatLon (geoTransform, reverseTrans, x, y);
        [cursorGeoTF setLatitude:y];
        [cursorGeoTF setLongitude:x];
    }
}

- (void) mouseDownOnImage: (NSPoint) point {
    int idx = [[pointMx selectedCell] tag];
    
    // Show point on map
    [view setCoordinates: point forIndex:idx];
    gcps[idx].dfGCPPixel = point.x;
    gcps[idx].dfGCPLine = point.y;
    
    [self updateGeoRefState];
}

- (void) sheetDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo {
    if (returnCode == NSAlertDefaultReturn) {
        [self writeCalibration];
    }
}

@end
